import logging
import multiprocessing
import random
import warnings
from pathlib import Path

import anndata as ad
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from skimage.morphology import disk, erosion
from tqdm import tqdm

from ..modules import bin_generation as bg
from ..modules.g4x_viewer import CellMasksSchema_pb2 as CellMasksSchema
from .decorator import workflow


@workflow
def seg_converter(
    adata: ad.AnnData,
    seg_mask: np.ndarray,
    out_path: str | Path,
    *,
    metadata: str | Path | None = None,
    cluster_key: str | None = None,
    emb_key: str | None = None,
    protein_list: list[str] | None = None,
    n_threads: int = 4,
    logger: logging.Logger | None = None,
) -> None:
    logger.info('Creating G4X-Viewer bin file.')

    warnings.filterwarnings(
        'ignore',
        message='FNV hashing is not implemented in Numba',
        category=UserWarning,
        module='numba.cpython.old_hashing',
    )

    if metadata is None:
        if cluster_key in adata.obs.columns:
            clusters_available = True
        else:
            clusters_available = False
        obs_df = adata.obs.copy()
    else:
        clusters_available = True
        clustered_df = pd.read_csv(metadata, index_col=0, header=0)
        if clustered_df.shape[1] > 1:
            assert cluster_key is not None, (
                'ERROR: multiple columns detected in cluster_info, cluster_key must be provided.'
            )
        else:
            cluster_key = clustered_df.columns[0]
        orig_df = adata.obs.copy()

        ## these are cells that were filtered out during clustering
        orig_df = orig_df.loc[list(set(orig_df.index) - set(clustered_df.index)), :].copy()
        for col in list(set(clustered_df.columns) - set(orig_df.columns)):
            orig_df[col] = '-1'

        obs_df = pd.concat([clustered_df, orig_df])
        obs_df.sort_index(inplace=True)

    ## initialize segmentation data
    ## we create polygons to define the boundaries of each cell mask
    logger.debug('Making polygons.')
    border = bg.get_border(seg_mask)
    seg_mask[border > 0] = 0
    eroded_mask = erosion(seg_mask, disk(1))
    outlines = seg_mask - eroded_mask
    sparse_matrix = csr_matrix(outlines)
    del seg_mask, border, eroded_mask, outlines

    nonzero_values = sparse_matrix.data
    nonzero_row_indices, nonzero_col_indices = sparse_matrix.nonzero()
    sorted_indices = np.argsort(nonzero_values)
    sorted_nonzero_values = nonzero_values[sorted_indices]
    sorted_rows = nonzero_row_indices[sorted_indices]
    sorted_cols = nonzero_col_indices[sorted_indices]

    ## add single-cell info
    logger.debug('Adding single-cell metadata.')
    cell_ids = obs_df.index.tolist()
    num_cells = len(cell_ids)
    centroid_y = obs_df['cell_x'].tolist()
    centroid_x = obs_df['cell_y'].tolist()

    if 'area' in obs_df.columns:
        areas = obs_df['area'].tolist()
    else:
        areas = obs_df['nuclei_expanded_area'].tolist()
    total_counts = obs_df['total_counts'].tolist()
    total_genes = obs_df['n_genes_by_counts'].tolist()

    if clusters_available:
        clusters = obs_df[cluster_key].tolist()
        cluster_palette = bg.generate_cluster_palette(clusters)
        cluster_colors = obs_df[cluster_key].astype(str).map(cluster_palette).tolist()
    else:
        cmap = plt.get_cmap('hsv', 100)
        clusters = ['-1'] * num_cells
        cluster_colors = [[int(255 * x) for x in cmap(i % 100)[:3]] for i in range(num_cells)]
        random.shuffle(cluster_colors)

    if protein_list:
        prot_vals = list(obs_df[protein_list].to_dict(orient='index').values())
    else:
        prot_vals = [{} for _ in range(num_cells)]
    if emb_key:
        umap_x = obs_df[f'{emb_key}_1'].to_numpy()
        umap_y = obs_df[f'{emb_key}_2'].to_numpy()
    else:
        umap_x = np.zeros(num_cells)
        umap_y = np.zeros(num_cells)

    ## refine polygons
    logger.debug('Refining polygons.')

    pq_args = [
        (k, cx, cy, sorted_nonzero_values, sorted_rows, sorted_cols)
        for k, cx, cy in zip(np.arange(1, num_cells + 1), centroid_x, centroid_y)
    ]

    with multiprocessing.Pool(processes=n_threads) as pool:
        polygons = pool.starmap(bg.refine_polygon, pq_args)

    ## do conversion
    # logger.debug(f"{sample_id}: Converting data to protobuff format...")
    segmentation_source = {
        'xs': [[y[0] for y in x] for x in polygons],
        'ys': [[y[1] for y in x] for x in polygons],
        'colors': cluster_colors,
        'cell_id': cell_ids,
        'area': areas,
        'centroid_x': centroid_x,
        'centroid_y': centroid_y,
        'total_tx': total_counts,
        'total_genes': total_genes,
        'cluster_id': clusters,
        'umap_x': umap_x,
        'umap_y': umap_y,
        'protein': prot_vals,
    }

    outputCellSegmentation = CellMasksSchema.CellMasks()

    for index in tqdm(range(len(segmentation_source['cell_id'])), desc='Processing cells'):
        try:
            cellPolygonPoints = [
                coord
                for pair in zip(segmentation_source['ys'][index], segmentation_source['xs'][index])
                for coord in pair
            ]
            cellPolygonColor = segmentation_source['colors'][index]
            cellId = segmentation_source['cell_id'][index]
            cellTotalCounts = segmentation_source['total_tx'][index]
            cellTotalGenes = segmentation_source['total_genes'][index]
            cellArea = segmentation_source['area'][index]
            clusterId = segmentation_source['cluster_id'][index]
            cellProt = segmentation_source['protein'][index]
            cellUmapX = segmentation_source['umap_x'][index]
            cellUmapY = segmentation_source['umap_y'][index]
        except Exception as e:
            logger.debug(e)
            pass
        outputMaskData = outputCellSegmentation.cellMasks.add()
        outputMaskData.vertices.extend(cellPolygonPoints + cellPolygonPoints[:2])
        outputMaskData.color.extend(cellPolygonColor)
        outputMaskData.cellId = str(cellId)
        outputMaskData.area = str(cellArea)
        outputMaskData.totalCounts = str(cellTotalCounts)
        outputMaskData.totalGenes = str(cellTotalGenes)
        outputMaskData.clusterId = str(clusterId)
        outputMaskData.umapValues.umapX = cellUmapX
        outputMaskData.umapValues.umapY = cellUmapY
        outputMaskData.proteins.update(cellProt)

    if clusters_available:
        for cluster_id, color in cluster_palette.items():
            entry = CellMasksSchema.ColormapEntry()
            entry.clusterId = cluster_id
            entry.color.extend(color)
            outputCellSegmentation.colormap.append(entry)
    else:
        entry = CellMasksSchema.ColormapEntry()
        entry.clusterId = '-1'
        entry.color.extend([int(31), int(119), int(180)])
        outputCellSegmentation.colormap.append(entry)

    ## write to file
    with open(out_path, 'wb') as file:
        file.write(outputCellSegmentation.SerializeToString())

    logger.debug(f'G4X-Viewer bin --> {out_path}')



import json
import math

# import re
import shutil
from datetime import datetime

# from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import polars as pl
from tqdm import tqdm

from .. import utils

if TYPE_CHECKING:
    from ..models import G4Xoutput

BASE_ORDER = 'CTGA'
LUT = np.zeros((256, 4), dtype=np.float32)
LUT[ord('C'), 0] = 1.0
LUT[ord('T'), 1] = 1.0
LUT[ord('G'), 2] = 1.0
LUT[ord('A'), 3] = 1.0

primer_read_map = {
    'SP1': 1,
    'm7a': 2,
    'm9a': 3,
    'm6a': 4,
    'm8a': 5,
    'm3a': 6,
}

PROBE_PATTERN = r'^(.*?)-([ACGT]{2,30})-([^-]+)$'


def one_hot_encode_str_array(seqs: list[str], seq_len: int) -> np.ndarray:
    """
    Fast one-hot encoding using LUT.
    Returns: (N, seq_len * 4) float32 array
    """
    N = len(seqs)
    # Flatten all sequences into a byte array and reshape to (N, seq_len)
    arr = np.frombuffer(''.join(seqs).encode('ascii'), dtype=np.uint8).reshape(N, seq_len)
    # Apply LUT: arr → (N, seq_len, 4), then flatten
    return LUT[arr].reshape(N, seq_len * 4)


def batched_dot_product_hamming_matrix(
    reads: list[str],
    codebook: list[str],
    batch_size: int,
) -> np.ndarray:
    """
    Compute full Hamming distance matrix (N_reads, N_codebook)
    using batched dot-product with one-hot encoding.
    """
    seq_len = len(codebook[0])
    assert all(len(seq) == seq_len for seq in codebook), 'All codebook entries must be same length'

    # One-hot encode the codebook once
    codebook_oh = one_hot_encode_str_array(codebook, seq_len)
    M = len(codebook)

    # Prepare final result
    N = len(reads)
    hamming_matrix = np.empty((N, M), dtype=np.uint8)

    num_expected_batches = math.ceil(N / batch_size)
    for i in tqdm(range(0, N, batch_size), total=num_expected_batches, desc='Demuxing batch', position=1, leave=False):
        batch_reads = reads[i : i + batch_size]
        batch_oh = one_hot_encode_str_array(batch_reads, seq_len)
        matches = batch_oh @ codebook_oh.T
        hamming = seq_len - matches
        hamming_matrix[i : i + len(batch_reads)] = hamming

    return hamming_matrix


def demux(
    hammings: np.ndarray,
    reads: pl.DataFrame,
    codebook_target_ids: np.ndarray,
    probe_dict: dict,
    max_ham_dist: int = 2,
    min_delta: int = 2,
) -> pl.DataFrame:
    demuxed = np.zeros(hammings.shape[0], dtype=bool)

    for i in range(max_ham_dist + 1):
        hits = hammings == i
        close_hits = hammings <= (i + min_delta)
        uniquely_hit = hits.sum(axis=1) == 1
        close_hit = close_hits.sum(axis=1) > 1
        pass_filter = uniquely_hit & ~close_hit
        demuxed[pass_filter] = 1

        # logger.info(f"""
        # ... ... {fq.stem}
        # hamming == {i}, min_delta == {min_delta}
        # unique hits = {sum(uniquely_hit)}
        # total cumulative hits within min_delta = {sum(close_hit)}
        # total demuxed (unique hits without another hit within min_delta) = {sum(pass_filter)}
        # """)

    # --- Get best hits ---
    hit_ids = hammings.argmin(axis=1)
    hit_targets = codebook_target_ids[hit_ids]

    transcripts = np.where(demuxed, hit_targets, 'UNDETERMINED')
    transcript_condensed = [probe_dict.get(t, 'UNDETERMINED') for t in transcripts]

    reads = reads.with_columns(
        [
            pl.Series('transcript_new', transcripts),
            pl.Series('transcript_condensed_new', transcript_condensed),
            pl.Series('demuxed_new', demuxed),
        ]
    )

    return reads


# def load_manifest(file_path: str | Path) -> tuple[pl.DataFrame, dict]:
#     manifest = pl.read_csv(file_path)
#     target_list = manifest['probe_name'].to_list()

#     p = re.compile('-[ACGT]{2,30}')
#     match = re.findall(pattern=p, string=target_list[0])
#     if len(match) == 0:
#         probe_dict = {p: '-'.join(p.split('-')[:-1]) for p in target_list}
#     else:
#         probe_dict = {}
#         for probe in target_list:
#             match = re.findall(pattern=p, string=probe)
#             probe_dict[probe] = probe.split(match[0])[0]
#     probe_dict['UNDETERMINED'] = 'UNDETERMINED'

#     return manifest, probe_dict


def parse_input_manifest(file_path: str) -> pl.DataFrame:
    """
    Parse strings in `df[col]` of the form '<prefix>-<15mer>-<primer>'
    and add columns: gene_name, sequence, primer, primer_code, read.
    Rows that don't match the pattern get nulls in the new columns.
    """
    df = pl.read_csv(file_path)

    col = 'probe_name'
    if col not in df.columns:
        raise ValueError(f"transcript manifest must contain column '{col}'")

    parsed = df.with_columns(
        [
            pl.col(col).str.extract(PROBE_PATTERN, 1).alias('gene_name'),
            pl.col(col).str.extract(PROBE_PATTERN, 2).alias('sequence'),
            pl.col(col).str.extract(PROBE_PATTERN, 3).alias('primer'),
        ]
    )

    null_count = parsed.null_count()['sequence'][0]
    if null_count > 0:
        print(f'{null_count} probes with invalid sequence format will be ignored:')
        null_seqs = parsed.filter(pl.col('sequence').is_null())['probe_name'].to_list()
        for ns in null_seqs:
            print(f'- {ns}')

    if 'read' in df.columns:
        print("Using 'read' column provided by input manifest.")
        parsed = parsed.drop('read')
    else:
        plist = parsed['primer'].unique().to_list()
        ign_primer = [p for p in plist if p not in primer_read_map]
        if len(ign_primer) > 0:
            print('Warning: the following primer names are not known and will be ignored:')
            for ip in ign_primer:
                print(f'- {ip}')
        parsed = parsed.filter(pl.col('primer').is_in(primer_read_map.keys())).with_columns(
            pl.col('primer').replace(primer_read_map).cast(pl.Int8).alias('read')
        )

    if 'gene_name' in df.columns:
        print("Using 'gene_name' column provided by input manifest.")
        parsed = parsed.drop('gene_name')

    manifest = parsed.join(df, on=col, how='left')

    probe_dict = dict(zip(manifest['probe_name'].to_list(), manifest['gene_name'].to_list()))
    probe_dict['UNDETERMINED'] = 'UNDETERMINED'

    return manifest, probe_dict


def update_metadata_and_tx_file(g4x_obj: 'G4Xoutput', manifest, out_dir):
    ## update metadata and transcript panel file
    shutil.copy(manifest, out_dir / 'transcript_panel.csv')

    with open(out_dir / 'run_meta.json', 'r') as f:
        meta = json.load(f)

    meta['transcript_panel'] = manifest.name
    meta['redemuxed_timestamp'] = f'{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}'
    (out_dir / 'run_meta.json').unlink()

    with open(out_dir / 'run_meta.json', 'w') as f:
        _ = json.dump(meta, f)
    meta = {'run_metadata': meta}
    (out_dir / 'g4x_viewer' / f'{g4x_obj.sample_id}_run_metadata.json').unlink()

    with open(out_dir / 'g4x_viewer' / f'{g4x_obj.sample_id}_run_metadata.json', 'w') as f:
        _ = json.dump(meta, f)


def batched_demuxing(g4x_obj: 'G4Xoutput', manifest, probe_dict, batch_dir, batch_size):
    seq_reads = manifest['read'].unique().to_list()
    seq_reads = [int(x.split('_')[-1]) if isinstance(x, str) else x for x in seq_reads]

    num_features = pl.scan_parquet(g4x_obj.feature_table_path).select(pl.len()).collect().item()
    num_expected_batches = math.ceil(num_features / batch_size)
    cols_to_select = [
        'x_coord_shift',
        'y_coord_shift',
        'z',
        'demuxed',
        'transcript_condensed',
        'meanQS',
        'sequence_to_demux',
        'transcript',
        'TXUID',
    ]
    for i, feature_batch in tqdm(
        enumerate(g4x_obj.stream_features(batch_size, cols_to_select)),
        total=num_expected_batches,
        desc='Demuxing transcripts',
        position=0,
    ):
        feature_batch = feature_batch.with_columns(pl.col('TXUID').str.split('_').list.last().cast(int).alias('read'))
        redemuxed_feature_batch = []
        for seq_read in seq_reads:
            feature_batch_read = feature_batch.filter(pl.col('read') == seq_read)
            manifest_read = manifest.filter(pl.col('read') == seq_read)

            if len(feature_batch_read) == 0 or len(manifest_read) == 0:
                continue

            seqs = feature_batch_read['sequence_to_demux'].to_list()
            codes = manifest_read['sequence'].to_list()
            codebook_target_ids = np.array(manifest_read['probe_name'].to_list())

            hammings = batched_dot_product_hamming_matrix(seqs, codes, batch_size=batch_size)
            feature_batch_read = demux(hammings, feature_batch_read, codebook_target_ids, probe_dict)
            redemuxed_feature_batch.append(feature_batch_read)

        pl.concat(redemuxed_feature_batch).write_parquet(batch_dir / f'batch_{i}.parquet')


def concatenate_and_cleanup(batch_dir, out_dir):
    final_tx_table_path = out_dir / 'rna' / 'transcript_table.csv'
    final_tx_pq_path = out_dir / 'diagnostics' / 'transcript_table.parquet'

    utils.delete_existing(final_tx_table_path)
    utils.delete_existing(final_tx_table_path.with_suffix('.csv.gz'))
    utils.delete_existing(final_tx_pq_path)

    tx_table = pl.scan_parquet(list(batch_dir.glob('*.parquet')))
    tx_table.sink_parquet(final_tx_pq_path)

    _ = (
        tx_table.filter(pl.col('demuxed_new'))
        .drop(
            'transcript',
            'transcript_new',
            'transcript_condensed',
            'demuxed',
            'demuxed_new',
            'sequence_to_demux',
            'TXUID',
        )
        .rename(
            {
                'transcript_condensed_new': 'gene_name',
                'x_coord_shift': 'y_pixel_coordinate',
                'y_coord_shift': 'x_pixel_coordinate',
                'z': 'z_level',
                'meanQS': 'confidence_score',
            }
        )
        .sink_csv(final_tx_table_path)
    )

    _ = utils.gzip_file(final_tx_table_path)
    shutil.rmtree(batch_dir)



from __future__ import annotations

import gzip
import logging
import os
import shutil
import traceback
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path

import rich_click as click
from rich.console import Console

console = Console()

DEFAULT_THREADS = max(1, (os.cpu_count() // 2 or 4))


def validate_data_dir(data_dir, resolve_path=False) -> Path:
    """check that expected outputs are present."""

    data_dir = validate_path(path_str=data_dir, must_exist=True, is_file_ok=False, resolve_path=resolve_path)

    # TODO: add more required files to check for
    required_paths = [
        'run_meta.json',
        'single_cell_data/feature_matrix.h5',
        'segmentation/segmentation_mask.npz',
        'rna/transcript_table.csv.gz',
    ]

    for p in required_paths:
        validate_path(data_dir / p, must_exist=True, is_dir_ok=False, is_file_ok=True)

    return data_dir


@contextmanager
def _spinner(message: str):
    with console.status(message, spinner='dots', spinner_style='red'):
        yield


def validate_path(path_str, must_exist=True, is_dir_ok=True, is_file_ok=True, resolve_path=False) -> Path:
    path = Path(path_str)  # .expanduser().resolve()
    if resolve_path:
        path = path.resolve()

    if must_exist and not path.exists():
        raise FileNotFoundError(f'Path does not exist: {path}')

    if path.exists():
        if path.is_dir() and not is_dir_ok:
            raise ValueError(f'Expected a file but got a directory: {path}')
        if path.is_file() and not is_file_ok:
            raise ValueError(f'Expected a directory but got a file: {path}')

    return path


def _fail_message(func_name, e, trace_back=False):
    click.echo('')
    click.secho(f'Failed {func_name}:', fg='red', err=True, bold=True)
    if trace_back:
        traceback.print_exc()
    raise click.ClickException(f'{type(e).__name__}: {e}')


def initialize_sample(
    data_dir: str, output: str | None, sample_id: str | None, n_threads: int = DEFAULT_THREADS
) -> None:
    msg = f'loading G4X-data from [blue]{data_dir}[/blue]'
    with _spinner(msg):
        import glymur

        from .models import G4Xoutput

        glymur.set_option('lib.num_threads', n_threads)
        try:
            sample = G4Xoutput(data_dir=data_dir, sample_id=sample_id)
        except Exception as e:
            click.echo('\n')
            click.secho('Failed to load G4X-data:', fg='red', err=True, bold=True)
            raise click.ClickException(f'{e}')

    if not output:
        click.secho('No output directory specified. Editing in-place!', fg='blue', bold=True)
        output = sample.data_dir

    return sample, output


# region file operations
def delete_existing(outfile: str | Path) -> None:
    outfile = Path(outfile)
    if outfile.exists() or outfile.is_symlink():
        outfile.unlink()


def gzip_file(outfile: str | Path, remove_original: bool = False) -> None:
    outfile = Path(outfile)
    delete_existing(outfile.with_suffix('.csv.gz'))
    with open(outfile, 'rb') as f_in:
        with gzip.open(outfile.with_suffix('.csv.gz'), 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    if remove_original:
        os.remove(outfile)


def create_custom_out(
    out_dir: Path | str,
    parent_folder: Path | str | None = None,
    file_name: str | None = None,
) -> Path:
    custom_out = Path(out_dir) / parent_folder

    # Ensure the directory exists
    custom_out.mkdir(parents=True, exist_ok=True)

    # Prepare output file path
    outfile = custom_out / file_name

    delete_existing(outfile)

    return outfile


# region logger
def verbose_to_log_level(verbose: int) -> int:
    """
    returns a logging level based on verbose integer:
    0 == WARNING
    1 == REPORT
    2 == INFO
    any other integer == DEBUG
    """
    mapping = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG,
    }
    return mapping.get(verbose, logging.DEBUG)


def setup_logger(
    logger_name: str,
    *,
    stream_logger: bool = True,
    stream_level: int = 2,
    file_logger: bool = True,
    file_level: int = 2,
    file_mode: str = 'a',
    out_dir: Path | str | None = None,
    format: str | None = None,
    clear_handlers: bool = True,
) -> logging.Logger:
    """
    Sets up a logger with configurable stream and file handlers.

    Parameters
    ----------
    logger_name : str
        Name of the logger.
    stream_logger : bool, optional
        Whether to enable logging to the console (default is True).
    stream_level : int, optional
        Logging level for the stream handler (default is logging.DEBUG).
    file_logger : bool, optional
        Whether to enable logging to a file (default is False).
    file_level : int, optional
        Logging level for the file handler (default is logging.DEBUG).
    file_mode : str, optional
        File mode for the log file. Common options: 'a' for append, 'w' for overwrite (default is 'a').
    out_dir : Path or str, optional
        Directory where the log file will be saved. Required if `file_logger` is True.
    format : str, optional
        Custom log message format. If not provided, a default format will be used.
    clear_handlers : bool, optional
        Whether to clear existing handlers from the logger before adding new ones (default is False).

    Returns
    -------
    logging.Logger
        Configured logger instance.

    """

    stream_level = verbose_to_log_level(stream_level)
    file_level = verbose_to_log_level(file_level)

    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)

    if format is None:
        format = '[%(asctime)s | %(name)s | %(levelname)s: %(message)s'

    formatter = logging.Formatter(format, datefmt='%Y-%m-%d %H:%M:%S')

    ## optionally clear existing handlers
    if clear_handlers:
        logger.handlers.clear()

    if stream_logger:
        h = logging.StreamHandler()
        h.setLevel(stream_level)
        h.setFormatter(formatter)
        logger.addHandler(h)

    if file_logger:
        assert out_dir is not None, 'out_dir must be provided if file_logger is True'
        out_dir = Path(out_dir)
        out_dir.mkdir(parents=True, exist_ok=True)

        timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
        log_path = out_dir / f'{logger_name}_{timestamp}.log'

        prior_size = log_path.stat().st_size if log_path.exists() else 0
        if file_mode == 'w':
            prior_size = 0

        fh = logging.FileHandler(log_path, mode=file_mode, encoding='utf-8')
        fh.setLevel(file_level)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

        session_lines = ['log session created']
        if prior_size > 0:
            session_lines.insert(0, '')
        fh.stream.write('\n'.join(session_lines) + '\n')
        fh.flush()

    return logger


def print_k_v(item, value, gap=2):
    value = '<undefined>' if not value else value
    click.secho(f'{item:<{gap}}', dim=True, nl=False)
    click.secho('- ', dim=True, nl=False)
    click.secho(f'{value}', fg='blue', bold=True)


def symlink_original_files(g4x_obj, out_dir: Path | str) -> None:
    ignore_file_list = ['clustering_umap.csv.gz', 'dgex.csv.gz', 'transcript_panel.csv']
    data_dir = Path(g4x_obj.data_dir).resolve()

    for root, dirs, files in os.walk(data_dir):
        rel_root = Path(root).relative_to(data_dir)
        if str(rel_root) == 'metrics':
            continue
        dst_root = out_dir / rel_root
        dst_root.mkdir(parents=True, exist_ok=True)

        for f in files:
            if f in ignore_file_list:
                continue
            # src_file = Path(root) / f
            dst_file = dst_root / f
            if dst_file.exists():
                dst_file.unlink()
            dst_file.symlink_to((Path(root) / f).resolve())



import atexit
import json
import logging
import shutil
import signal
import sys
import tarfile
from pathlib import Path

from .decorator import workflow


@workflow
def tar_viewer(
    viewer_dir: Path,
    out_dir: Path,
    *,
    logger: logging.Logger,
):
    logger.info('Checking files.')
    viewer_dir = Path(viewer_dir)
    assert viewer_dir.exists(), f'{viewer_dir} does not appear to exist.'

    bin_path = list(viewer_dir.glob('*.bin'))
    assert len(bin_path) == 1, 'Either no bin file was found in viewer_dir or multiple bin files were found.'
    bin_path = bin_path[0]

    sample_id = bin_path.stem

    ome_tiff_path = viewer_dir / f'{sample_id}.ome.tiff'
    assert ome_tiff_path.is_file(), 'ome.tiff file does not exist.'

    run_meta_path = viewer_dir / f'{sample_id}_run_metadata.json'
    assert run_meta_path.is_file(), 'run_metadata.json file does not exist.'

    tx_path = viewer_dir / f'{sample_id}.tar'
    # assert tx_path.is_file(), 'transcript tar file does not exist.'

    # --- H&E paths
    h_and_e_path = viewer_dir / 'h_and_e'
    created_he_dir = False
    if not h_and_e_path.exists():
        h_and_e_path.mkdir(parents=True, exist_ok=True)
        created_he_dir = True

    orig_he = viewer_dir / f'{sample_id}_HE.ome.tiff'
    assert orig_he.is_file(), 'fH&E ome.tiff file does not exist.'
    moved_he = h_and_e_path / f'{sample_id}_HE.ome.tiff'

    moved = False
    restored = False

    def restore_he():
        nonlocal moved, restored, created_he_dir
        if restored:
            return
        try:
            if moved and moved_he.exists():
                if orig_he.exists():
                    orig_he.unlink()
                shutil.move(str(moved_he), str(orig_he))
                logger.info('Restored H&E file to original location.')

            # NEW: remove the h_and_e folder if we created it and it’s empty
            try:
                if created_he_dir and h_and_e_path.exists() and not any(h_and_e_path.iterdir()):
                    h_and_e_path.rmdir()
                    logger.info('Removed empty h_and_e directory.')
            except Exception as e:
                logger.warning(f'failed to remove h_and_e directory: {e}')

            restored = True
            moved = False
        except Exception as e:
            logger.warning(f'failed to restore H&E file: {e}')

    atexit.register(restore_he)

    def _handle_signal(sig):
        import signal as _signal

        sig_name = _signal.Signals(sig).name
        logger.warning(f'Received {sig_name}; cleaning up…')
        restore_he()
        sys.exit(128 + sig)

    signal.signal(signal.SIGINT, _handle_signal)
    signal.signal(signal.SIGTERM, _handle_signal)

    # Move H&E into subfolder for packaging
    shutil.move(str(orig_he), str(moved_he))
    moved = True

    try:
        logger.info('Making metadata.')
        metadata = {
            'protein_image_src': ome_tiff_path.name,
            'protein_image_data_src': run_meta_path.name,
            'he_images_src': h_and_e_path.name,
            'cell_segmentation_src': bin_path.name,
            'transcript_src': tx_path.name,
        }

        with open(viewer_dir / 'dataset.config.json', 'w') as f:
            json.dump(metadata, f)

        logger.info('Tarring folder.')
        out_tar = Path(out_dir) / 'g4x_viewer'

        if not out_tar.exists():
            out_tar.mkdir(parents=True, exist_ok=True)

        out_tar = out_tar / f'{sample_id}.tar'

        with tarfile.open(out_tar, 'w') as tar:
            tar.add(viewer_dir, arcname=viewer_dir.name)

    except Exception as e:
        logger.error(f'Workflow failed: {e}')
        raise

    finally:
        logger.info('Restoring moved files...')
        restore_he()
        try:
            atexit.unregister(restore_he)
        except Exception:
            pass



import inspect

import rich_click as click

from . import __version__, utils

# click.rich_click.THEME = 'modern'
click.rich_click.MAX_WIDTH = 100
click.rich_click.COMMANDS_PANEL_TITLE = 'commands'
click.rich_click.OPTIONS_PANEL_TITLE = 'options'
click.rich_click.STYLE_OPTION = 'bold blue'
click.rich_click.STYLE_ARGUMENT = 'bold blue'
click.rich_click.STYLE_COMMAND = 'bold blue'
click.rich_click.STYLE_SWITCH = 'bold red'
click.rich_click.STYLE_METAVAR = 'bold red'
click.rich_click.STYLE_METAVAR_SEPARATOR = 'dim'
click.rich_click.STYLE_USAGE = 'bold yellow'
click.rich_click.STYLE_USAGE_COMMAND = 'bold'
click.rich_click.STYLE_HELPTEXT_FIRST_LINE = ''
click.rich_click.STYLE_HELPTEXT = 'dim'
click.rich_click.STYLE_OPTION_DEFAULT = 'dim'
click.rich_click.STYLE_REQUIRED_SHORT = 'bold yellow'
click.rich_click.STYLE_REQUIRED_LONG = 'bold yellow'
click.rich_click.STYLE_OPTIONS_PANEL_BORDER = 'dim'
click.rich_click.STYLE_COMMANDS_PANEL_BORDER = 'dim'
click.rich_click.COMMANDS_BEFORE_OPTIONS = True

# click.rich_click.ARGUMENTS_PANEL_TITLE = 'in/out'

click.rich_click.COMMAND_GROUPS = {
    'g4x-helpers': [
        {'name': 'commands', 'commands': ['redemux', 'resegment', 'update_bin', 'new_bin', 'tar_viewer']},
        # {"name": "utilities", "commands": ["log"]},
    ],
}


# click.rich_click.OPTION_GROUPS = {
#     'g4x-helpers': [
#         {
#             'name': 'in/out',  #
#             'options': ['--input', '--output'],
#         },
#         {
#             'name': 'options',  #
#             'options': [
#                 '--sample-id',
#                 '--threads',
#                 '--verbose',
#                 '--version',
#                 '--help',
#             ],
#         },
#     ]
# }

CLI_HELP = 'Helper models and post-processing tools for G4X-data\n\ndocs.singulargenomics.com'

RESEG_HELP = (
    'Reprocess G4X-output with a new segmentation\n\n'
    'Takes custom cell segmentation labels as input and re-assigns transcripts and protein signals to those cells.'
)

REDMX_HELP = 'Reprocess G4X-output with a new transcript manifest'

UDBIN_HELP = 'Update existing G4X-viewer .bin file with new metadata'

NWBIN_HELP = 'Create new G4X-viewer .bin file from sample directory'

TARVW_HELP = (
    'Package G4X-viewer folder for distribution\n\n'
    'Creates a .tar archive of the "g4x_viewer" directory for easy upload and sharing.\n\n'
    'Archive file name: {SAMPLE_ID}_g4x_viewer.tar\n\n'
)

VALIDATE_HELP = 'Validate G4X-data directory structure and files'


def io_opts(*, takes_input=False, writes_output=True):
    options = []

    def decorator(f):
        if takes_input:
            options.append(
                click.option(
                    '-i',
                    '--g4x-data',
                    required=True,
                    type=click.Path(exists=True, file_okay=False),
                    help='Directory containing G4X-data for a single sample',
                    panel='data i/o',
                )
            )
            options.append(
                click.option(
                    '-s',
                    '--sample-id',  #
                    required=False,
                    type=str,
                    help='Sample ID (used for naming outputs)',
                    panel='data i/o',
                )
            )
        if writes_output:
            options.append(
                click.option(
                    '-o',
                    '--output',
                    required=False,
                    type=click.Path(exists=True, file_okay=False, dir_okay=True, writable=True),
                    help='Output directory used. Will edit G4X-data in-place if not provided.',
                    panel='data i/o',
                )
            )

        for opt in reversed(options):
            f = opt(f)
        return f

    return decorator


# region cli
@click.group(
    context_settings=dict(help_option_names=['-h', '--help']),
    invoke_without_command=True,
    add_help_option=True,
    help=CLI_HELP,
)
@click.option(
    '-t',
    '--threads',
    required=False,
    type=int,
    default=utils.DEFAULT_THREADS,
    show_default=True,
    help='Number of threads to use for processing',
)
@click.option(
    '-v',
    '--verbose',
    required=False,
    type=int,
    default=1,
    show_default=True,
    help='Console logging level (0, 1, 2)',
)
@click.option(
    '--version',
    is_flag=True,
    default=False,
    help='Display g4x-helpers version',
)
@click.pass_context
# @click.option('-v', '--verbose', count=True, help="Increase verbosity")
def cli(ctx, threads, verbose, version):
    if version:
        click.echo(f'g4x-helpers: {__version__}')
        ctx.exit()

    # No subcommand and no input given → show help
    if not ctx.invoked_subcommand:
        click.echo(ctx.get_help())
        ctx.exit()

    if ctx.invoked_subcommand:
        ctx.ensure_object(dict)
        ctx.obj['threads'] = threads
        ctx.obj['verbose'] = verbose
        ctx.obj['version'] = __version__


############################################################
# region resegment
@cli.command(name='resegment', help=RESEG_HELP)
@io_opts(takes_input=True, writes_output=True)
@click.option(
    '--cell-labels',
    required=True,
    type=click.Path(exists=True, dir_okay=False),
    help='File containing cell segmentation labels.\n\nsupported file types: [.npy, .npz, .geojson]',
)
@click.option(
    '--labels-key',
    required=False,
    type=str,
    default=None,
    help='Key/column in npz/geojson where labels should be taken from (optional)',
)
@click.pass_context
def cli_resegment(ctx, g4x_data, output, sample_id, cell_labels, labels_key):
    func_name = inspect.currentframe().f_code.co_name.removeprefix('cli_')

    g4x_obj, output = utils.initialize_sample(
        data_dir=g4x_data, output=output, sample_id=sample_id, n_threads=ctx.obj['threads']
    )

    try:
        with utils._spinner(f'Initializing {func_name} process...'):
            from .main_features import resegment

        resegment(
            g4x_obj=g4x_obj,
            out_dir=output,
            cell_labels=cell_labels,
            labels_key=labels_key,
            n_threads=ctx.obj['threads'],
            verbose=ctx.obj['verbose'],
        )
    except Exception as e:
        utils._fail_message(func_name, e, trace_back=False)


############################################################
# region redemux
@cli.command(name='redemux', help=REDMX_HELP)
@io_opts(takes_input=True, writes_output=True)
@click.option(
    '--manifest',
    required=True,
    type=click.Path(exists=True, dir_okay=False),
    help='Path to manifest for demuxing. The manifest must be a 3-column CSV with the following header: target,sequence,read.',
)
@click.option(
    '--batch-size',
    default=1_000_000,
    show_default=True,
    type=int,
    help='Number of transcripts to process per batch.',
)
@click.pass_context
def cli_redemux(ctx, g4x_data, output, sample_id, manifest, batch_size):
    func_name = inspect.currentframe().f_code.co_name.removeprefix('cli_')
    g4x_obj, output = utils.initialize_sample(
        data_dir=g4x_data, output=output, sample_id=sample_id, n_threads=ctx.obj['threads']
    )
    try:
        with utils._spinner(f'Initializing {func_name} process...'):
            from .main_features import redemux

        redemux(
            g4x_obj=g4x_obj,
            out_dir=output,
            manifest=manifest,
            batch_size=batch_size,
            n_threads=ctx.obj['threads'],
            verbose=ctx.obj['verbose'],
        )
    except Exception as e:
        utils._fail_message(func_name, e)


############################################################
# region update_bin
@cli.command(name='update_bin', help=UDBIN_HELP)
@io_opts(takes_input=True, writes_output=True)
@click.option(
    '--metadata',
    required=True,
    type=click.Path(exists=True, dir_okay=False),
    help='Path to metadata table where clustering and/or embedding information will be extracted. Table must contain a header.',
)
@click.option(
    '--bin-file',
    type=click.Path(exists=True, dir_okay=False),
    help='Path to a specific .bin file to update.',
)
@click.option(
    '--cellid-key',
    default=None,
    type=str,
    help='Column name in metadata containing cell IDs that match with bin_file. If not provided, assumes that first column in metadata contains the cell IDs.',
)
@click.option(
    '--cluster-key',
    default=None,
    type=str,
    help='Column name in metadata containing cluster IDs. Automatically assigns new colors if cluster_color_key is not provided. If not provided, skips updating cluster IDs.',
)
@click.option(
    '--cluster-color-key',
    default=None,
    type=str,
    help='Column name in metadata containing cluster colors. Colors must be provided as hex codes. If provided, cluster_key must also be provided. If not provided, skips updating cluster colors.',
)
@click.option(
    '--emb-key',
    default=None,
    type=str,
    help='Column name in metadata containing embedding. Parser will look for {emb_key}_1 and {emb_key}_2. If not provided, skips updating embedding.',
)
@click.pass_context
def cli_update_bin(
    ctx, g4x_data, output, sample_id, metadata, bin_file, cellid_key, cluster_key, cluster_color_key, emb_key
):
    func_name = inspect.currentframe().f_code.co_name.removeprefix('cli_')
    g4x_obj, output = utils.initialize_sample(
        data_dir=g4x_data, output=output, sample_id=sample_id, n_threads=ctx.obj['threads']
    )
    try:
        with utils._spinner(f'Initializing {func_name} process...'):
            from .main_features import update_bin

        update_bin(
            g4x_obj=g4x_obj,
            out_dir=output,
            bin_file=bin_file,
            metadata=metadata,
            cellid_key=cellid_key,
            cluster_key=cluster_key,
            cluster_color_key=cluster_color_key,
            emb_key=emb_key,
            verbose=ctx.obj['verbose'],
        )
    except Exception as e:
        utils._fail_message(func_name, e)


############################################################
# region new_bin
@cli.command(name='new_bin', help=NWBIN_HELP)
@io_opts(takes_input=True, writes_output=True)
@click.pass_context
def cli_new_bin(ctx, g4x_data, output, sample_id):
    func_name = inspect.currentframe().f_code.co_name.removeprefix('cli_')
    g4x_obj, output = utils.initialize_sample(
        data_dir=g4x_data, output=output, sample_id=sample_id, n_threads=ctx.obj['threads']
    )
    try:
        with utils._spinner(f'Initializing {func_name} process...'):
            from .main_features import new_bin

        new_bin(
            g4x_obj=g4x_obj,  #
            out_dir=output,
            n_threads=ctx.obj['threads'],
            verbose=ctx.obj['verbose'],
        )
    except Exception as e:
        utils._fail_message(func_name, e, trace_back=True)


############################################################
# region tar_viewer
@cli.command(name='tar_viewer', help=TARVW_HELP)
@io_opts(takes_input=True, writes_output=True)
@click.option(
    '--viewer-dir',
    required=False,
    type=click.Path(exists=True, file_okay=False),
    help='(optional) Path to G4X-viewer folder. If set, will archive the specified folder instead of the one supplied by G4X-data.',
)
@click.pass_context
def cli_tar_viewer(ctx, g4x_data, output, sample_id, viewer_dir):
    func_name = inspect.currentframe().f_code.co_name.removeprefix('cli_')
    g4x_obj, output = utils.initialize_sample(
        data_dir=g4x_data, output=output, sample_id=sample_id, n_threads=ctx.obj['threads']
    )
    try:
        with utils._spinner(f'Initializing {func_name} process...'):
            from .main_features import tar_viewer

        tar_viewer(
            g4x_obj=g4x_obj,
            out_dir=output,
            viewer_dir=viewer_dir,
            verbose=ctx.obj['verbose'],
        )
    except Exception as e:
        utils._fail_message(func_name, e)


if __name__ == '__main__':
    cli(prog_name='g4x-helpers')



from __future__ import annotations

import logging
import math
import multiprocessing as mp
import os
import shutil
from pathlib import Path
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import polars as pl

from ..modules import tx_generation as tg
from .decorator import workflow

if TYPE_CHECKING:
    from ..models import G4Xoutput

mp.set_start_method('spawn', force=True)


@workflow
def tx_converter(
    g4x_obj: 'G4Xoutput',
    out_path: Path,
    *,
    aggregation_level: str = 'gene',
    n_threads: int = 4,
    sampling_fraction: float = 0.2,
    logger: logging.Logger,
) -> None:
    logger.info('Generating viewer transcript file.')

    ## prelims
    IMAGE_RESOLUTION = g4x_obj.shape
    if g4x_obj.platform == 'g4x-2lane':
        MIN_TILE_SIZE = 1028
    else:
        MIN_TILE_SIZE = 512
    out_dir = g4x_obj.data_dir / 'g4x_viewer_temp'
    os.makedirs(out_dir, exist_ok=True)

    ## get transcript table
    if aggregation_level == 'probe':
        tx_column = 'transcript'
    else:
        tx_column = 'gene_name'
    keep_cols = ['x_pixel_coordinate', 'y_pixel_coordinate', 'cell_id', tx_column]
    df = g4x_obj.load_transcript_table(lazy=True, columns=keep_cols)

    ## make colormap
    unique_genes = df.select(tx_column).collect().unique().to_series().to_list()
    num_genes = len(unique_genes)
    base_cmap = plt.get_cmap('hsv', num_genes)
    palette = {gene: [int(255 * c) for c in base_cmap(i / num_genes)[:3]] for i, gene in enumerate(unique_genes)}

    df = df.with_columns(
        (pl.col('x_pixel_coordinate') / MIN_TILE_SIZE).cast(pl.Int32).alias('tile_y_coord'),
        (pl.col('y_pixel_coordinate') / MIN_TILE_SIZE).cast(pl.Int32).alias('tile_x_coord'),
        pl.col(tx_column).replace_strict(palette, default=[127, 127, 127]).alias('color'),
        pl.concat_list(['x_pixel_coordinate', 'y_pixel_coordinate']).alias('position'),
    )

    num_tiles_x, num_tiles_y = tg.GetPyramidTilesConfigData(IMAGE_RESOLUTION, MIN_TILE_SIZE)
    NUMBER_OF_LEVELS = tg.GetPyramidLevelsConfigData(IMAGE_RESOLUTION, MIN_TILE_SIZE)

    min_zoom_tiles_x = math.ceil(num_tiles_x / (pow(2, NUMBER_OF_LEVELS - 1)))
    min_zoom_tiles_y = math.ceil(num_tiles_y / (pow(2, NUMBER_OF_LEVELS - 1)))

    logger.info(f"""
        Final configurations for parsing:
            Image resolution: {IMAGE_RESOLUTION[0]} x {IMAGE_RESOLUTION[1]}
            Number of max zoom tiles: X = {num_tiles_x} | Y = {num_tiles_y}
            Number of min zoom tiles: X = {min_zoom_tiles_x} | Y = {min_zoom_tiles_y}
            Number of levels: {NUMBER_OF_LEVELS}
    """)

    tg.save_configuration_file(out_dir, IMAGE_RESOLUTION, MIN_TILE_SIZE, NUMBER_OF_LEVELS, palette)
    logger.info('Parsing and classifying tiles...')

    pq_args = []
    for level_index in reversed(range(NUMBER_OF_LEVELS + 1)):
        ## subsampling factor
        sampling_factor = sampling_fraction ** (NUMBER_OF_LEVELS - level_index)

        # factor for computing tile coordinates at this level
        scaling_factor = 2 ** (NUMBER_OF_LEVELS - level_index)
        current_tile_size = MIN_TILE_SIZE * scaling_factor
        x_num_of_tiles = math.ceil(IMAGE_RESOLUTION[0] / current_tile_size)
        y_num_of_tiles = math.ceil(IMAGE_RESOLUTION[1] / current_tile_size)

        # Ensure even numbers of tiles
        if x_num_of_tiles % 2 != 0:
            x_num_of_tiles += 1
        if y_num_of_tiles % 2 != 0:
            y_num_of_tiles += 1

        for tile_x_index in range(x_num_of_tiles):
            tileOutputDirPath = out_dir / f'{level_index}' / f'{tile_x_index}'
            os.makedirs(tileOutputDirPath, exist_ok=True)
            pq_args.append([tileOutputDirPath, y_num_of_tiles, scaling_factor])

            _ = (
                df.filter(
                    ((pl.col('tile_x_coord') // scaling_factor) == tile_x_index),
                )
                .select(['position', 'color', tx_column, 'cell_id', 'tile_y_coord'])
                .collect()
                .sample(fraction=sampling_factor)
                .write_parquet(tileOutputDirPath / 'tmp.parquet')
            )

    ctx = mp.get_context('spawn')
    with ctx.Pool(processes=n_threads) as pool:
        pool.starmap(tg._process_x, pq_args)

    logger.info('Tarring up.')
    if out_path.exists() or out_path.is_symlink():
        out_path.unlink()
    _ = tg.create_tar_from_directory(out_dir, out_path)
    shutil.rmtree(out_dir, ignore_errors=True)



from collections import deque
from pathlib import Path

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
from numba import njit
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import minimum_spanning_tree
from scipy.spatial import distance_matrix
from skimage.measure import approximate_polygon
from skimage.morphology import dilation, disk

from .. import utils


def bin_file_path(g4x_obj, out_dir: str | Path | None = None) -> Path:
    file_name = f'{g4x_obj.sample_id}.bin'

    if out_dir:
        out_dir = utils.validate_path(out_dir, must_exist=True, is_dir_ok=True, is_file_ok=False)
        outfile = Path(out_dir) / file_name
    else:
        outfile = utils.create_custom_out(out_dir, 'g4x_viewer', file_name)

    return outfile


def generate_cluster_palette(clusters: list, max_colors: int = 256) -> dict:
    """
    Generate a color palette mapping for cluster labels.

    This function assigns RGB colors to unique cluster labels using a matplotlib colormap.
    Clusters labeled as "-1" are assigned a default gray color `[191, 191, 191]`.

    The colormap used depends on the number of clusters:
        - `tab10` for ≤10 clusters
        - `tab20` for ≤20 clusters
        - `hsv` for more than 20 clusters, capped by `max_colors`

    Parameters
    ----------
    clusters : list
        A list of cluster identifiers (strings or integers). The special label '-1' is excluded
        from color mapping and handled separately.
    max_colors : int, optional
        Maximum number of colors to use in the HSV colormap. Only used if there are more than
        20 unique clusters. Default is 256.

    Returns
    -------
    dict
        A dictionary mapping each cluster ID (as a string) to a list of three integers
        representing an RGB color in the range [0, 255].

    Examples
    --------
    >>> generate_cluster_palette(['0', '1', '2', '-1'])
    {'0': [31, 119, 180], '1': [255, 127, 14], '2': [44, 160, 44], '-1': [191, 191, 191]}
    """
    unique_clusters = [c for c in np.unique(clusters) if c != '-1']
    n_clusters = len(unique_clusters)

    if n_clusters <= 10:
        base_cmap = plt.get_cmap('tab10')
    elif n_clusters <= 20:
        base_cmap = plt.get_cmap('tab20')
    else:
        base_cmap = plt.get_cmap('hsv', min(max_colors, n_clusters))

    cluster_palette = {
        str(cluster): [int(255 * c) for c in base_cmap(i / n_clusters)[:3]] for i, cluster in enumerate(unique_clusters)
    }
    cluster_palette['-1'] = [int(191), int(191), int(191)]

    return cluster_palette


def hex2rgb(hex: str) -> list[int, int, int]:
    return [int(x * 255) for x in mcolors.to_rgb(hex)]


@njit
def get_start_stop_idx(arr, k):
    start_idx = np.searchsorted(arr, k, side='left')
    end_idx = np.searchsorted(arr, k, side='right')
    return start_idx, end_idx


def returnEndpoints(adj_list, adjacency=2):
    # Identify endpoints of the MST
    endpoints = [node for node in adj_list if len(adj_list[node]) == adjacency]

    return endpoints


def bfs_path(start, end, adj_list):
    queue = deque([(start, [start])])
    visited = set()
    while queue:
        current, path = queue.popleft()
        if current == end:
            return path
        if current in visited:
            continue
        visited.add(current)
        for neighbor in adj_list[current]:
            if neighbor not in visited:
                queue.append((neighbor, path + [neighbor]))
    return []


def computeLongestPath(adj_list):
    endpoints = returnEndpoints(adj_list)
    longest_path = []
    max_length = 0
    # Use a dictionary to cache paths and avoid recomputation
    path_cache = {}
    # Compute distances between all pairs of endpoints
    for i in range(len(endpoints)):
        for j in range(i + 1, len(endpoints)):
            if (endpoints[i], endpoints[j]) not in path_cache:
                path = bfs_path(endpoints[i], endpoints[j], adj_list)
                path_cache[(endpoints[i], endpoints[j])] = path
            else:
                path = path_cache[(endpoints[i], endpoints[j])]
            if len(path) > max_length:
                max_length = len(path)
                longest_path = path

    return longest_path


@njit
def createAdjacencyList_numba(mst):
    """
    Create an adjacency list from a minimum spanning tree (MST) using Numba for performance optimization.

    Parameters:
    mst (numpy.ndarray): The minimum spanning tree represented as a 2D numpy array.

    Returns:
    tuple: A tuple containing:
        - adj_list (numpy.ndarray): An array where each row contains the adjacent nodes for each node.
        - adj_list_pos (numpy.ndarray): An array containing the number of adjacent nodes for each node.
    """
    n = mst.shape[0]
    adj_list = np.zeros((n, n * 2), dtype=np.uint32)
    adj_list_pos = np.zeros(n, dtype=np.uint32)
    for i in range(n):
        for j in range(n):
            if mst[i, j] != 0 or mst[j, i] != 0:
                adj_list[i, adj_list_pos[i]] = j
                adj_list_pos[i] += 1
                adj_list[j, adj_list_pos[j]] = i
                adj_list_pos[j] += 1

    return adj_list, adj_list_pos


def indicesToArray(points, longest_path):
    pth = []

    for j in range(len(longest_path)):
        pth.append([points[longest_path[j], 0], points[longest_path[j], 1]])

    return np.array(pth)


def simplify_polygon(points: np.ndarray, tolerance: float) -> np.ndarray:
    """
    Simplify a series of points representing a polygon using scikit-image's
    approximate_polygon. The tolerance controls how aggressively the polygon
    is simplified (in pixel units).
    """
    if len(points) <= 2:
        return points

    # If the first and last points are not the same, append the first to the end
    # to ensure the polygon is "closed" for approximation (optional).
    if not np.array_equal(points[0], points[-1]):
        points = np.vstack([points, points[0]])

    # Perform polygon simplification
    simplified = approximate_polygon(points, tolerance=tolerance)

    # If approximate_polygon returns only the closed ring, remove the last point
    # to avoid duplication in your pipeline. (approx_polygon returns a closed ring
    # by repeating the first point at the end.)
    if len(simplified) > 2 and np.array_equal(simplified[0], simplified[-1]):
        simplified = simplified[:-1]

    return simplified


def pointsToSingleSmoothPath(points: np.ndarray, tolerance: float) -> np.ndarray:
    # Calculate the distance matrix
    dist_matrix = distance_matrix(points, points)

    # Create a sparse matrix for the MST calculation
    sparse_matrix = csr_matrix(dist_matrix)

    # Compute the MST
    mst = minimum_spanning_tree(sparse_matrix).toarray()

    adj_list, adj_list_pos = createAdjacencyList_numba(mst)
    adj_list = {row: list(adj_list[row, :pos]) for row, pos in enumerate(adj_list_pos) if pos}

    longest_path = computeLongestPath(adj_list)
    bestPath = indicesToArray(points, longest_path)

    simplified_path = simplify_polygon(bestPath, tolerance=tolerance)

    return simplified_path


def refine_polygon(k, cx, cy, sorted_nonzero_values_ref, sorted_rows_ref, sorted_cols_ref):
    start_idx, end_idx = get_start_stop_idx(sorted_nonzero_values_ref, k)
    points = np.vstack((sorted_rows_ref[start_idx:end_idx], sorted_cols_ref[start_idx:end_idx])).T
    return pointsToSingleSmoothPath(points, tolerance=2.0)


def get_border(mask: np.ndarray, s: int = 1) -> np.ndarray:
    d = dilation(mask, disk(s))
    border = (mask != d).astype(np.uint8)
    return border



import functools

from ..utils import setup_logger


def workflow(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        logger = kwargs.pop('logger', None)
        if logger is None:
            logger = setup_logger(logger_name=func.__name__, file_logger=False)
            logger.info('No logger provided, using default logger (stream only).')

        logger.info('-' * 10)
        logger.info(f'Initializing {func.__name__} workflow.')

        result = func(*args, logger=logger, **kwargs)

        logger.info(f'Completed {func.__name__} workflow.')
        return result

    return wrapper



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import anndata as ad
import geopandas
import numpy as np
import polars as pl
import scanpy as sc
import shapely.affinity
import skimage.measure
from geopandas import GeoDataFrame
from rasterio.features import rasterize
from rasterio.transform import Affine
from shapely.geometry import Polygon
from skimage.measure._regionprops import RegionProperties
from tqdm import tqdm

if TYPE_CHECKING:
    from geopandas.geodataframe import GeoDataFrame

    from g4x_helpers.models import G4Xoutput

SUPPORTED_MASK_FILETYPES = {'.npy', '.npz', '.geojson'}


def try_load_segmentation(
    cell_labels: Path, expected_shape: tuple[int], labels_key: str | None = None
) -> np.ndarray | geopandas.GeoDataFrame:
    ## load new segmentation
    cell_labels = Path(cell_labels)
    suffix = cell_labels.suffix.lower()

    if cell_labels.suffix not in SUPPORTED_MASK_FILETYPES:
        raise ValueError(f'{cell_labels.suffix} is not a supported file type.')

    if suffix == '.npz':
        with np.load(cell_labels) as labels:
            available_keys = list(labels.keys())

            if labels_key:  # if a key is specified
                if labels_key not in labels:
                    raise KeyError(f"Key '{labels_key}' not found in .npz; available keys: {available_keys}")
                seg = labels[labels_key]

            else:  # if no key specified,
                if len(labels) == 1:  # and only one key is available, use that key
                    seg = labels[available_keys[0]]
                else:  # and multiple keys are available, raise an error
                    if len(labels) > 1:
                        raise ValueError(
                            f"Multiple keys found in .npz: {available_keys}.\nPlease specify a key using 'labels_key'"
                        )
                    seg = labels

    elif suffix == '.npy':
        # .npy: directly returns the array, no context manager available
        if labels_key is not None:
            print('file is .npy, ignoring provided labels_key.')
        seg = np.load(cell_labels)

    elif suffix == '.geojson':
        seg = geopandas.read_file(cell_labels)

        if labels_key is not None:
            if labels_key not in seg.columns:
                raise KeyError(f"Column '{labels_key}' not found in GeoJSON; available columns: {seg.columns.tolist()}")

            # ensure that a coliumn named 'label' exists
            seg['label'] = seg[labels_key]

        else:
            if 'label' not in seg.columns:
                raise ValueError(
                    "No column named 'label' found in GeoJSON. Please specify which column to use for labels via labels_key."
                )

    # only validate shape for numpy arrays
    if isinstance(seg, np.ndarray):
        assert seg.shape == expected_shape, (
            f'provided mask shape {seg.shape} does not match G4X sample shape {expected_shape}'
        )

    return seg


def _make_cell_metadata(segmentation_props: pl.DataFrame, cell_by_protein: pl.DataFrame) -> pl.DataFrame:
    cell_metadata = segmentation_props.join(cell_by_protein, on='cell_id', how='left')
    return cell_metadata


def _make_cell_by_gene(segmentation_props: pl.DataFrame, reads_new_labels: pl.DataFrame) -> pl.DataFrame:
    cell_by_gene = (
        reads_new_labels.filter(pl.col('segmentation_label') != 0)
        .group_by('cell_id', 'gene_name')
        .agg(pl.len().alias('counts'))
        .sort('gene_name')
        .pivot(on='gene_name', values='counts', index='cell_id')
    )

    cell_by_gene = segmentation_props.select('cell_id').join(cell_by_gene, on='cell_id', how='left').fill_null(0)
    return cell_by_gene


def _make_adata(cell_by_gene: pl.DataFrame, cell_metadata: pl.DataFrame) -> ad.AnnData:
    X = cell_by_gene.drop('cell_id').to_numpy()

    obs_df = cell_metadata.drop('segmentation_label').to_pandas()
    obs_df.index = obs_df.index.astype(str)

    adata = ad.AnnData(X=X, obs=obs_df)

    adata.var['gene_id'] = cell_by_gene.columns[1:]
    adata.var['modality'] = 'tx'
    adata.obs_names = adata.obs['cell_id']
    adata.var_names = adata.var['gene_id']

    sc.pp.calculate_qc_metrics(adata, inplace=True, percent_top=None)

    return adata


def get_mask_properties(sample: 'G4Xoutput', mask: np.ndarray) -> pl.DataFrame:
    props = skimage.measure.regionprops(mask)

    prop_dict = []
    # Loop through each region to get the area and centroid, with a progress bar
    for prop in tqdm(props, desc='Extracting mask properties'):
        label = prop.label  # The label (mask id)
        area = prop.area  # Area: count of pixels
        centroid = prop.centroid  # Centroid: (row, col)

        # assuming coordinate order: 'yx':
        cell_y, cell_x = centroid

        prop_dict.append(
            {
                'segmentation_label': label,
                'cell_id': f'{sample.sample_id}-{label}',
                'area': area,
                'cell_x': cell_x,
                'cell_y': cell_y,
            }
        )

    prop_dict_df = pl.DataFrame(prop_dict)
    return prop_dict_df


def assign_tx_to_mask_labels(sample: 'G4Xoutput', mask: np.ndarray) -> pl.DataFrame:
    reads = sample.load_transcript_table()

    coord_order = ['y_pixel_coordinate', 'x_pixel_coordinate']

    tx_coords = reads[coord_order].to_numpy()
    cell_ids = mask[tx_coords[:, 0].astype(int), tx_coords[:, 1].astype(int)]

    reads = reads.with_columns(pl.lit(cell_ids).alias('segmentation_label'))
    reads = reads.with_columns((f'{sample.sample_id}-' + pl.col('segmentation_label').cast(pl.String)).alias('cell_id'))

    return reads


def image_mask_intensity_extraction(
    mask: np.ndarray,
    img: np.ndarray,
    *,
    bead_mask: np.ndarray | None = None,
    label_prefix: str = '',
    channel_label: str = 'intensity_mean',
    lazy: bool = True,
) -> pl.DataFrame | pl.LazyFrame:
    mask_flat = mask.ravel()
    img_flat = img.ravel()

    ## optional masking with beads
    if bead_mask is not None:
        bead_mask_flat = ~bead_mask.ravel()
        mask_flat = mask_flat[bead_mask_flat]
        img_flat = img_flat[bead_mask_flat]

    # intensity_means = np.bincount(mask_flat, weights=img_flat)[1:] / np.bincount(mask_flat)[1:]
    counts = np.bincount(mask_flat)[1:]
    sums = np.bincount(mask_flat, weights=img_flat)[1:]

    # Avoid division by zero:
    intensity_means = np.divide(sums, counts, out=np.zeros_like(sums), where=counts != 0)

    if lazy:
        prop_tab = pl.LazyFrame(
            {'label': np.arange(start=1, stop=mask_flat.max() + 1), channel_label: intensity_means}
        ).with_columns((pl.lit(label_prefix) + pl.col('label').cast(pl.Utf8)).cast(pl.Categorical).alias('label'))
    else:
        prop_tab = pl.DataFrame(
            {'label': np.arange(start=1, stop=mask_flat.max() + 1), channel_label: intensity_means}
        ).with_columns((pl.lit(label_prefix) + pl.col('label').cast(pl.Utf8)).cast(pl.Categorical).alias('label'))

    return prop_tab


def join_frames(lf_left, lf_right):
    return lf_left.join(lf_right, on='label', how='inner')


def _region_props_to_polygons(region_props: RegionProperties) -> list[Polygon]:
    mask = np.pad(region_props.image, 1)
    contours = skimage.measure.find_contours(mask, 0.5)

    # shapes with <= 3 vertices, i.e. lines, can't be converted into a polygon
    polygons = [Polygon(contour[:, [1, 0]]) for contour in contours if contour.shape[0] >= 4]

    yoff, xoff, *_ = region_props.bbox
    return [shapely.affinity.translate(poly, xoff, yoff) for poly in polygons]


def _vectorize_mask(mask: np.ndarray, nudge: bool = True) -> GeoDataFrame:
    if mask.max() == 0:
        return GeoDataFrame(geometry=[])

    regions = skimage.measure.regionprops(mask)

    geoms = []
    labels = []
    # Wrap the iteration in tqdm to show progress
    for region in tqdm(regions, desc='Vectorizing regions'):
        polys = _region_props_to_polygons(region)
        geoms.extend(polys)
        # add the region label once per polygon
        labels.extend([region.label] * len(polys))

    gdf = GeoDataFrame({'label': labels}, geometry=geoms)

    if nudge:
        # GeoSeries.translate works elementwise
        gdf['geometry'] = gdf['geometry'].translate(xoff=-0.5, yoff=-0.5)

    return gdf


def rasterize_polygons(gdf: GeoDataFrame, target_shape: tuple) -> np.ndarray:
    height, width = target_shape
    transform = Affine.identity()

    # wrap the zip in tqdm; total=len(gdf) gives a proper progress bar
    wrapped = tqdm(zip(gdf.geometry, gdf['label']), total=len(gdf), desc='Rasterizing polygons')
    # feed that wrapped iterator into rasterize
    shapes = ((geom, int(lbl)) for geom, lbl in wrapped)

    label_array = rasterize(
        shapes=shapes,
        out_shape=(height, width),
        transform=transform,
        fill=0,  # background value
        dtype='int32',
    )

    return label_array


def extract_image_signals(
    sample: G4Xoutput, mask: np.ndarray, lazy: bool = False, signal_list: list[str] | None = None, cached: bool = False
) -> pl.DataFrame | pl.LazyFrame:
    if signal_list is None:
        signal_list = ['nuclear', 'eosin'] + sample.proteins

    channel_name_map = {protein: protein for protein in sample.proteins}
    channel_name_map['nuclear'] = 'nuclearstain'
    channel_name_map['eosin'] = 'cytoplasmicstain'

    for signal_name in tqdm(signal_list, desc='Extracting protein signal'):
        if signal_name not in ['nuclear', 'eosin']:
            image_type = 'protein'
            protein = signal_name
        else:
            image_type = signal_name
            protein = None

        signal_img = sample.load_image_by_type(image_type, thumbnail=False, protein=protein, cached=cached)

        ch_label = f'{channel_name_map[signal_name]}_intensity_mean'

        prop_tab = image_mask_intensity_extraction(
            mask,
            signal_img,
            bead_mask=None,
            label_prefix=f'{sample.sample_id}-',
            channel_label=ch_label,
            lazy=lazy,
        )

        if signal_name == signal_list[0]:
            signal_df = prop_tab
        else:
            signal_df = join_frames(signal_df, prop_tab)

    signal_df = signal_df.cast({'label': pl.String}).rename({'label': 'cell_id'})

    return signal_df



from .redemux import redemux as redemux
from .seg_converter import seg_converter as seg_converter
from .seg_intersect import intersect_segmentation as intersect_segmentation
from .seg_updater import seg_updater as seg_updater
from .tar_viewer import tar_viewer as tar_viewer
from .tx_converter import tx_converter as tx_converter



import functools
import logging
from typing import TYPE_CHECKING

import rich_click as click

from . import __version__, utils
from . import workflows as wflw

if TYPE_CHECKING:
    from .models import G4Xoutput


def _base_command(func):
    """Decorator to apply standard command initialization logic."""

    @functools.wraps(func)
    def wrapper(
        g4x_obj: 'G4Xoutput',
        out_dir: str,
        n_threads: int = utils.DEFAULT_THREADS,
        verbose: int = 1,
        file_logger: bool = True,
        **kwargs,
    ):
        func_name = func.__name__

        out_dir = utils.validate_path(out_dir, must_exist=False, is_dir_ok=True, is_file_ok=False)

        if out_dir != g4x_obj.data_dir:
            func_out = out_dir / func_name
            func_out.mkdir(parents=True, exist_ok=True)
            out_dir = func_out

        gap = 12
        click.secho(f'\nStarting: {func_name}\n', bold=True)
        utils.print_k_v('sample_dir', f'{g4x_obj.data_dir}', gap)
        utils.print_k_v('out_dir', f'{out_dir}', gap)
        utils.print_k_v('n_threads', f'{n_threads}', gap)
        utils.print_k_v('verbosity', f'{verbose}', gap)
        utils.print_k_v('g4x-helpers', f'v{__version__}', gap)
        click.echo('')

        if not kwargs.get('logger', None):
            logger = utils.setup_logger(
                logger_name=func_name, out_dir=out_dir / 'logs', stream_level=verbose, file_logger=file_logger
            )
        logger.info(f'Running {func_name} with G4X-helpers v{__version__}')

        # Pass the initialized logger and parameters to the wrapped function
        result = func(
            g4x_obj=g4x_obj,
            out_dir=out_dir,
            n_threads=n_threads,
            verbose=verbose,
            logger=logger,
            **kwargs,
        )

        click.secho(f'\nCompleted: {func.__name__}', bold=True, fg='green')
        return result

    return wrapper


@_base_command
def resegment(
    g4x_obj: 'G4Xoutput',
    out_dir: str,
    cell_labels: str,
    *,
    labels_key: str | None = None,
    n_threads: int = 4,
    logger: logging.Logger,
    **kwargs,
) -> None:
    from .modules import segmentation as seg

    cell_labels = utils.validate_path(cell_labels, must_exist=True, is_dir_ok=False, is_file_ok=True)

    labels = seg.try_load_segmentation(
        cell_labels=cell_labels,
        expected_shape=g4x_obj.shape,
        labels_key=labels_key,
    )

    adata, mask = wflw.intersect_segmentation(
        g4x_obj=g4x_obj,
        labels=labels,
        out_dir=out_dir,
        logger=logger,
    )

    # TODO this pattern is repeated in several places, refactor into utility function
    view_dir = out_dir / 'g4x_viewer'
    view_dir.mkdir(parents=True, exist_ok=True)

    wflw.seg_converter(
        adata=adata,
        seg_mask=mask,
        out_path=view_dir / f'{g4x_obj.sample_id}.bin',
        protein_list=[f'{x}_intensity_mean' for x in g4x_obj.proteins],
        n_threads=n_threads,
        logger=logger,
    )


@_base_command
def redemux(
    g4x_obj: 'G4Xoutput',
    out_dir: str,
    manifest: str,
    *,
    batch_size: int = 1_000_000,
    n_threads: int = 4,
    logger: logging.Logger,
    **kwargs,
) -> None:
    wflw.redemux(
        g4x_obj=g4x_obj,
        manifest=manifest,
        out_dir=out_dir,
        batch_size=batch_size,
        logger=logger,
    )

    ## now regenerate the secondary files
    logger.info('Regenerating downstream files.')
    g4x_obj.data_dir = out_dir  ## set base directory to the redemux output folder for downstream steps

    # resegment with existing segmentation
    logger.info('Intersecting with existing cell segmentation.')
    adata, mask = wflw.intersect_segmentation(
        g4x_obj=g4x_obj,
        labels=g4x_obj.load_segmentation(),
        out_dir=out_dir,
        logger=logger,
    )

    view_dir = out_dir / 'g4x_viewer'
    view_dir.mkdir(parents=True, exist_ok=True)

    wflw.seg_converter(
        adata=adata,
        seg_mask=mask,
        out_path=view_dir / f'{g4x_obj.sample_id}.bin',
        protein_list=[f'{x}_intensity_mean' for x in g4x_obj.proteins],
        n_threads=n_threads,
        logger=logger,
    )

    wflw.tx_converter(
        g4x_obj=g4x_obj,
        out_path=view_dir / f'{g4x_obj.sample_id}.tar',
        n_threads=n_threads,
        logger=logger,
    )


@_base_command
def update_bin(
    g4x_obj: 'G4Xoutput',
    out_dir: str,
    metadata: str,
    *,
    bin_file: str | None = None,
    cellid_key: str | None = None,
    cluster_key: str | None = None,
    cluster_color_key: str | None = None,
    emb_key: str | None = None,
    logger: logging.Logger,
    **kwargs,
) -> None:
    metadata = utils.validate_path(metadata, must_exist=True, is_dir_ok=False, is_file_ok=True)

    view_dir = out_dir / 'g4x_viewer'
    view_dir.mkdir(parents=True, exist_ok=True)

    wflw.seg_updater(
        bin_file=g4x_obj.data_dir / 'g4x_viewer' / f'{g4x_obj.sample_id}.bin' if bin_file is None else bin_file,
        metadata_file=metadata,
        out_path=view_dir / f'{g4x_obj.sample_id}.bin',
        cellid_key=cellid_key,
        cluster_key=cluster_key,
        cluster_color_key=cluster_color_key,
        emb_key=emb_key,
        logger=logger,
    )


@_base_command
def new_bin(
    g4x_obj: 'G4Xoutput',  #
    out_dir: str,
    *,
    n_threads: int = 4,
    logger: logging.Logger,
    **kwargs,
) -> None:
    ## set up the data
    try:
        adata = g4x_obj.load_adata()
        emb_key = '_'.join(sorted([x for x in adata.obs.columns if 'X_umap' in x])[0].split('_')[:-1])
        cluster_key = sorted([x for x in adata.obs.columns if 'leiden' in x])[0]
        logger.info('Successfully loaded adata with clustering information.')

    except Exception:
        adata = g4x_obj.load_adata(load_clustering=False)
        emb_key = None
        cluster_key = None
        logger.info('Clustering information was not found, cell coloring will be random.')

    view_dir = out_dir / 'g4x_viewer'
    view_dir.mkdir(parents=True, exist_ok=True)

    wflw.seg_converter(
        adata=adata,
        seg_mask=g4x_obj.load_segmentation(),
        out_path=view_dir / f'{g4x_obj.sample_id}.bin',
        cluster_key=cluster_key,
        emb_key=emb_key,
        protein_list=[f'{x}_intensity_mean' for x in g4x_obj.proteins],
        n_threads=n_threads,
        logger=logger,
    )


@_base_command
def tar_viewer(
    g4x_obj: 'G4Xoutput',  #
    out_dir: str | None = None,
    *,
    logger: logging.Logger,
    **kwargs,
) -> None:
    viewer_dir = g4x_obj.data_dir / 'g4x_viewer'
    wflw.tar_viewer(viewer_dir=viewer_dir, out_dir=out_dir, logger=logger)



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class UmapEntry(_message.Message):
    __slots__ = ('umapY', 'umapX')
    UMAPY_FIELD_NUMBER: _ClassVar[int]
    UMAPX_FIELD_NUMBER: _ClassVar[int]
    umapY: float
    umapX: float
    def __init__(self, umapY: _Optional[float] = ..., umapX: _Optional[float] = ...) -> None: ...

class SingleMask(_message.Message):
    __slots__ = (
        'vertices',
        'color',
        'area',
        'totalCounts',
        'totalGenes',
        'cellId',
        'clusterId',
        'proteins',
        'umapValues',
    )
    class ProteinsEntry(_message.Message):
        __slots__ = ('key', 'value')
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: float
        def __init__(self, key: _Optional[str] = ..., value: _Optional[float] = ...) -> None: ...

    VERTICES_FIELD_NUMBER: _ClassVar[int]
    COLOR_FIELD_NUMBER: _ClassVar[int]
    AREA_FIELD_NUMBER: _ClassVar[int]
    TOTALCOUNTS_FIELD_NUMBER: _ClassVar[int]
    TOTALGENES_FIELD_NUMBER: _ClassVar[int]
    CELLID_FIELD_NUMBER: _ClassVar[int]
    CLUSTERID_FIELD_NUMBER: _ClassVar[int]
    PROTEINS_FIELD_NUMBER: _ClassVar[int]
    UMAPVALUES_FIELD_NUMBER: _ClassVar[int]
    vertices: _containers.RepeatedScalarFieldContainer[float]
    color: _containers.RepeatedScalarFieldContainer[int]
    area: str
    totalCounts: str
    totalGenes: str
    cellId: str
    clusterId: str
    proteins: _containers.ScalarMap[str, float]
    umapValues: UmapEntry
    def __init__(
        self,
        vertices: _Optional[_Iterable[float]] = ...,
        color: _Optional[_Iterable[int]] = ...,
        area: _Optional[str] = ...,
        totalCounts: _Optional[str] = ...,
        totalGenes: _Optional[str] = ...,
        cellId: _Optional[str] = ...,
        clusterId: _Optional[str] = ...,
        proteins: _Optional[_Mapping[str, float]] = ...,
        umapValues: _Optional[_Union[UmapEntry, _Mapping]] = ...,
    ) -> None: ...

class ColormapEntry(_message.Message):
    __slots__ = ('clusterId', 'color')
    CLUSTERID_FIELD_NUMBER: _ClassVar[int]
    COLOR_FIELD_NUMBER: _ClassVar[int]
    clusterId: str
    color: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, clusterId: _Optional[str] = ..., color: _Optional[_Iterable[int]] = ...) -> None: ...

class CellMasks(_message.Message):
    __slots__ = ('cellMasks', 'colormap', 'numberOfCells')
    CELLMASKS_FIELD_NUMBER: _ClassVar[int]
    COLORMAP_FIELD_NUMBER: _ClassVar[int]
    NUMBEROFCELLS_FIELD_NUMBER: _ClassVar[int]
    cellMasks: _containers.RepeatedCompositeFieldContainer[SingleMask]
    colormap: _containers.RepeatedCompositeFieldContainer[ColormapEntry]
    numberOfCells: int
    def __init__(
        self,
        cellMasks: _Optional[_Iterable[_Union[SingleMask, _Mapping]]] = ...,
        colormap: _Optional[_Iterable[_Union[ColormapEntry, _Mapping]]] = ...,
        numberOfCells: _Optional[int] = ...,
    ) -> None: ...



from typing import (
    ClassVar as _ClassVar,
)
from typing import (
    Iterable as _Iterable,
)
from typing import (
    Mapping as _Mapping,
)
from typing import (
    Optional as _Optional,
)
from typing import (
    Union as _Union,
)

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf.internal import containers as _containers

DESCRIPTOR: _descriptor.FileDescriptor

class PointData(_message.Message):
    __slots__ = ('position', 'color', 'geneName', 'cellId')
    POSITION_FIELD_NUMBER: _ClassVar[int]
    COLOR_FIELD_NUMBER: _ClassVar[int]
    GENENAME_FIELD_NUMBER: _ClassVar[int]
    CELLID_FIELD_NUMBER: _ClassVar[int]
    position: _containers.RepeatedScalarFieldContainer[float]
    color: _containers.RepeatedScalarFieldContainer[int]
    geneName: str
    cellId: str
    def __init__(
        self,
        position: _Optional[_Iterable[float]] = ...,
        color: _Optional[_Iterable[int]] = ...,
        geneName: _Optional[str] = ...,
        cellId: _Optional[str] = ...,
    ) -> None: ...

class TileData(_message.Message):
    __slots__ = ('pointsData', 'numberOfPoints')
    POINTSDATA_FIELD_NUMBER: _ClassVar[int]
    NUMBEROFPOINTS_FIELD_NUMBER: _ClassVar[int]
    pointsData: _containers.RepeatedCompositeFieldContainer[PointData]
    numberOfPoints: int
    def __init__(
        self, pointsData: _Optional[_Iterable[_Union[PointData, _Mapping]]] = ..., numberOfPoints: _Optional[int] = ...
    ) -> None: ...

class ColumnData(_message.Message):
    __slots__ = ('columnTiles',)
    COLUMNTILES_FIELD_NUMBER: _ClassVar[int]
    columnTiles: _containers.RepeatedCompositeFieldContainer[TileData]
    def __init__(self, columnTiles: _Optional[_Iterable[_Union[TileData, _Mapping]]] = ...) -> None: ...

class LevelData(_message.Message):
    __slots__ = ('levelColumns',)
    LEVELCOLUMNS_FIELD_NUMBER: _ClassVar[int]
    levelColumns: _containers.RepeatedCompositeFieldContainer[ColumnData]
    def __init__(self, levelColumns: _Optional[_Iterable[_Union[ColumnData, _Mapping]]] = ...) -> None: ...

class Metadata(_message.Message):
    __slots__ = ('level',)
    LEVEL_FIELD_NUMBER: _ClassVar[int]
    level: _containers.RepeatedCompositeFieldContainer[LevelData]
    def __init__(self, level: _Optional[_Iterable[_Union[LevelData, _Mapping]]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: MetadataSchema.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x14MetadataSchema.proto"V\n\tPointData\x12\x14\n\x08position\x18\x01 \x03(\x01\x42\x02\x10\x01\x12\x11\n\x05\x63olor\x18\x02 \x03(\rB\x02\x10\x01\x12\x10\n\x08geneName\x18\x03 \x01(\t\x12\x0e\n\x06\x63\x65llId\x18\x04 \x01(\t"B\n\x08TileData\x12\x1e\n\npointsData\x18\x01 \x03(\x0b\x32\n.PointData\x12\x16\n\x0enumberOfPoints\x18\x02 \x01(\r",\n\nColumnData\x12\x1e\n\x0b\x63olumnTiles\x18\x01 \x03(\x0b\x32\t.TileData".\n\tLevelData\x12!\n\x0clevelColumns\x18\x01 \x03(\x0b\x32\x0b.ColumnData"%\n\x08Metadata\x12\x19\n\x05level\x18\x01 \x03(\x0b\x32\n.LevelDatab\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'MetadataSchema_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals['_POINTDATA'].fields_by_name['position']._loaded_options = None
    _globals['_POINTDATA'].fields_by_name['position']._serialized_options = b'\020\001'
    _globals['_POINTDATA'].fields_by_name['color']._loaded_options = None
    _globals['_POINTDATA'].fields_by_name['color']._serialized_options = b'\020\001'
    _globals['_POINTDATA']._serialized_start = 24
    _globals['_POINTDATA']._serialized_end = 110
    _globals['_TILEDATA']._serialized_start = 112
    _globals['_TILEDATA']._serialized_end = 178
    _globals['_COLUMNDATA']._serialized_start = 180
    _globals['_COLUMNDATA']._serialized_end = 224
    _globals['_LEVELDATA']._serialized_start = 226
    _globals['_LEVELDATA']._serialized_end = 272
    _globals['_METADATA']._serialized_start = 274
    _globals['_METADATA']._serialized_end = 311
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import json
import math
import os
import tarfile
from pathlib import Path

import polars as pl


def create_tar_from_directory(directory_path: str | Path, archive_name: str | Path) -> None:
    if not os.path.isdir(directory_path):
        raise ValueError(f'The directory {directory_path} does not exist.')
    with tarfile.open(archive_name, 'w') as tar:
        tar.add(directory_path, arcname=os.path.basename(directory_path))


def GetPyramidTilesConfigData(image_resolution: tuple[int, int], tile_size: int) -> tuple[int, int]:
    num_tiles_x = math.ceil(image_resolution[0] / tile_size)
    num_tiles_y = math.ceil(image_resolution[1] / tile_size)

    if not num_tiles_x % 2 == 0:
        num_tiles_x += 1

    if not num_tiles_y % 2 == 0:
        num_tiles_y += 1

    return num_tiles_x, num_tiles_y


def GetPyramidLevelsConfigData(image_resolution: tuple[int, int], tile_size: int, min_tiles_number: int = 16) -> int:
    min_num_levels = 0

    current_level = 1
    while min_num_levels == 0:
        level_tile_size = tile_size * pow(2, current_level)
        level_tiles_x = math.ceil(image_resolution[0] / level_tile_size)
        level_tiles_y = math.ceil(image_resolution[1] / level_tile_size)
        if not level_tiles_x % 2 == 0:
            level_tiles_x += 1
        if not level_tiles_y % 2 == 0:
            level_tiles_y += 1
        if level_tiles_x * level_tiles_y <= min_tiles_number:
            min_num_levels = current_level
            break
        current_level += 1

    return min_num_levels


def save_configuration_file(
    outputDirPath: str, image_resolution: tuple[int, int], min_tile_size: int, number_of_levels: int, palette: dict
) -> None:
    start_tile_size = min_tile_size * pow(2, number_of_levels)

    config_data = {
        'layer_height': image_resolution[0],
        'layer_width': image_resolution[1],
        'layers': number_of_levels,
        'tile_size': start_tile_size,
        'color_map': [{'gene_name': key, 'color': value} for key, value in palette.items()],
    }

    with open(f'{outputDirPath}/config.json', 'w') as json_file:
        json.dump(config_data, json_file, indent=2)


def _process_x(tileOutputDirPath: Path, y_num_of_tiles: int, scaling_factor: int) -> None:
    from ..modules.g4x_viewer import MetadataSchema_pb2 as MetadataSchema

    for tile_y_index in range(y_num_of_tiles):
        outputTileData = MetadataSchema.TileData()

        df_current = (
            pl.scan_parquet(tileOutputDirPath / 'tmp.parquet')
            .filter(((pl.col('tile_y_coord') // scaling_factor) == tile_y_index))
            .drop('tile_y_coord')
            .collect()
        )
        # Iterate over rows directly
        ## this can potentially be done lazily with this PR: https://github.com/pola-rs/polars/pull/23980
        for position, color, gene, cell_id in df_current.iter_rows():
            outputPointData = outputTileData.pointsData.add()
            _ = outputPointData.position.extend(position)
            _ = outputPointData.color.extend(color)
            outputPointData.geneName = gene
            outputPointData.cellId = str(cell_id)

        with open(f'{tileOutputDirPath}/{tile_y_index}.bin', 'wb') as file:
            _ = file.write(outputTileData.SerializeToString())



import logging
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from geopandas import GeoDataFrame

from .. import utils
from ..modules import segmentation as seg
from .decorator import workflow

if TYPE_CHECKING:
    from ..models import G4Xoutput


@workflow
def intersect_segmentation(
    g4x_obj: 'G4Xoutput',
    labels: np.ndarray | GeoDataFrame,
    out_dir: Path,
    *,
    exclude_channels: list[str] | None = None,
    logger: logging.Logger,
) -> None:
    logger.info(f'Using provided output directory: {out_dir}')

    signal_list = ['nuclear', 'eosin'] + g4x_obj.proteins

    if exclude_channels is not None:
        logger.info(f'Not processing channels: {", ".join(exclude_channels)}')
        if isinstance(exclude_channels, str):
            exclude_channels = [exclude_channels]

        signal_list = [item for item in signal_list if item not in exclude_channels]
    else:
        logger.info('Processing all channels.')

    if isinstance(labels, GeoDataFrame):
        logger.info('Rasterizing provided GeoDataFrame.')
        mask = seg.rasterize_polygons(gdf=labels, target_shape=g4x_obj.shape)
    else:
        mask = labels

    logger.info('Extracting mask properties.')
    segmentation_props = seg.get_mask_properties(g4x_obj, mask)

    logger.info('Assigning transcripts to mask labels.')
    reads_new_labels = seg.assign_tx_to_mask_labels(g4x_obj, mask)

    logger.info('Extracting protein signal.')
    cell_by_protein = seg.extract_image_signals(g4x_obj, mask, signal_list=signal_list)

    logger.info('Building output data structures.')
    cell_metadata = seg._make_cell_metadata(segmentation_props, cell_by_protein)
    cell_by_gene = seg._make_cell_by_gene(segmentation_props, reads_new_labels)
    adata = seg._make_adata(cell_by_gene, cell_metadata)

    probe_types = g4x_obj.load_adata(remove_nontargeting=False, load_clustering=False).var['probe_type']
    adata.var = adata.var.merge(probe_types, left_index=True, right_index=True, how='left')

    logger.info(f'Saving output files to {out_dir}')

    outfile = utils.create_custom_out(out_dir, 'segmentation', 'segmentation_mask.npz')
    logger.debug(f'segmentation mask --> {outfile}')
    np.savez(outfile, cell_labels=mask)

    outfile = utils.create_custom_out(out_dir, 'rna', 'transcript_table.csv')
    logger.debug(f'transcript table --> {outfile}.gz')
    reads_new_labels.write_csv(outfile)
    _ = utils.gzip_file(outfile, remove_original=True)

    outfile = utils.create_custom_out(out_dir, 'single_cell_data', 'cell_by_transcript.csv')
    logger.debug(f'cell x transcript --> {outfile}.gz')
    cell_by_gene.write_csv(outfile)
    _ = utils.gzip_file(outfile, remove_original=True)

    outfile = utils.create_custom_out(out_dir, 'single_cell_data', 'cell_by_protein.csv')
    logger.debug(f'cell x protein --> {outfile}.gz')
    cell_by_protein.write_csv(outfile)
    _ = utils.gzip_file(outfile, remove_original=True)

    outfile = utils.create_custom_out(out_dir, 'single_cell_data', 'feature_matrix.h5')
    logger.debug(f'single-cell h5 --> {outfile}')
    adata.write_h5ad(outfile)

    outfile = utils.create_custom_out(out_dir, 'single_cell_data', 'cell_metadata.csv.gz')
    logger.debug(f'cell metadata --> {outfile}')
    adata.obs.to_csv(outfile, compression='gzip')

    return adata, mask



import logging
from pathlib import Path
from typing import TYPE_CHECKING

from .. import utils
from ..modules import demultiplexing as dmx
from .decorator import workflow

if TYPE_CHECKING:
    from ..models import G4Xoutput


@workflow
def redemux(
    g4x_obj: 'G4Xoutput',
    manifest: Path,
    out_dir: Path,
    *,
    batch_size: int = 1_000_000,
    logger: logging.Logger,
) -> None:
    ## preflight checks
    logger.info('Validating input paths.')
    manifest = utils.validate_path(manifest, must_exist=True, is_dir_ok=False, is_file_ok=True)
    out_dir = utils.validate_path(out_dir, must_exist=False, is_dir_ok=True, is_file_ok=False)

    batch_dir = out_dir / 'diagnostics' / 'batches'
    batch_dir.mkdir(parents=True, exist_ok=True)

    ## make output directory with symlinked files from original
    logger.info('Creating output directory and symlinking original files.')
    utils.symlink_original_files(g4x_obj, out_dir)

    ## update metadata and transcript panel file
    logger.info('Updating metadata and transcript panel file.')
    dmx.update_metadata_and_tx_file(g4x_obj, manifest, out_dir)

    ## load the new manifest file that we will demux against
    logger.info('Loading manifest file.')
    manifest, probe_dict = dmx.parse_input_manifest(manifest)

    logger.info('Performing re-demuxing.')
    dmx.batched_demuxing(g4x_obj, manifest, probe_dict, batch_dir, batch_size)

    ## concatenate results into final csv and parquet
    logger.info('Writing updated transcript table.')
    dmx.concatenate_and_cleanup(batch_dir, out_dir)



import logging
from pathlib import Path

import pandas as pd
from tqdm import tqdm

from ..modules import bin_generation as bg
from ..modules.g4x_viewer import CellMasksSchema_pb2 as CellMasksSchema
from .decorator import workflow


@workflow
def seg_updater(
    bin_file: Path,
    metadata_file: Path,
    out_path: Path,
    *,
    cellid_key: str | None = None,
    cluster_key: str | None = None,
    cluster_color_key: str | None = None,
    emb_key: str | None = None,
    logger: logging.Logger,
) -> None:
    ## pre-flight
    if emb_key is None and cluster_key is None:
        logger.warning('neither embedding nor cluster keys were provided, nothing to update.')
        return None

    ## load the bin file
    logger.info(f'Loading {bin_file}.')
    with open(bin_file, 'rb') as f:
        data = f.read()
    cell_masks = CellMasksSchema.CellMasks()
    cell_masks.ParseFromString(data)

    ## load the metadata
    if cellid_key is None:
        logger.info('cellid_key not provided, assuming cell IDs are in first column of metadata.')
        metadata = pd.read_csv(metadata_file, index_col=0, header=0)
    else:
        metadata = pd.read_csv(metadata_file, index_col=None, header=0)
        if cellid_key not in metadata.columns:
            raise KeyError(f'{cellid_key} not a valid column in metadata.')
        metadata.set_index(cellid_key, inplace=True)

    ## check for clustering
    if cluster_key is not None:
        if cluster_key not in metadata.columns:
            raise KeyError(f'{cluster_key} not a valid column in metadata.')
        update_cluster = True
        logger.debug('Updating cluster IDs.')
    else:
        update_cluster = False
        logger.debug('Not updating cluster IDs.')

    ## check for cluster colors
    if cluster_color_key is not None:
        if cluster_key is None:
            raise ValueError('cluster_color_key was provided, but cluster_key was not provided.')
        if cluster_color_key not in metadata.columns:
            raise KeyError(f'{cluster_color_key} not a valid column in metadata.')
        color = metadata[cluster_color_key].iat[0]
        assert color.startswith('#'), 'Cluster colors must be provided as hexcodes.'
        update_cluster_color = True
        logger.debug('Updating cluster colors.')
        cluster_palette = (
            metadata.drop_duplicates(subset=cluster_key)[[cluster_key, cluster_color_key]]
            .set_index(cluster_key)
            .to_dict()[cluster_color_key]
        )
        cluster_palette = {str(k): bg.hex2rgb(v) for k, v in cluster_palette.items()}
    else:
        if cluster_key is not None:
            update_cluster_color = True
            logger.debug('Auto-assigning colors to new clustering.')
            cluster_color_key = 'cluster_color'
            cluster_palette = bg.generate_cluster_palette(metadata[cluster_key].tolist())
            metadata['cluster_color'] = metadata[cluster_key].astype(str).map(cluster_palette).tolist()
        else:
            update_cluster_color = False
            logger.debug('Not updating cluster colors.')

    ## check for embedding
    if emb_key is not None:
        if f'{emb_key}_1' not in metadata.columns or f'{emb_key}_2' not in metadata.columns:
            raise KeyError(f'{emb_key}_1 and {emb_key}_2 are not valid columns in metadata.')
        update_emb = True
        logger.debug('Updating embedding.')
    else:
        update_emb = False
        logger.debug('Not updating embedding.')

    ## Do the actual updating
    logger.info('Updating cells.')
    for cell in tqdm(cell_masks.cellMasks, desc='Updating cell data'):
        current_cellid = cell.cellId
        if current_cellid in metadata.index:
            if update_cluster:
                cell.clusterId = str(metadata.loc[current_cellid, cluster_key])
            if update_cluster_color:
                # clear out the existing color entries:
                cell.ClearField('color')
                cell.color.extend(metadata.loc[current_cellid, cluster_color_key])
            if update_emb:
                cell.umapValues.umapX = metadata.loc[current_cellid, f'{emb_key}_1']
                cell.umapValues.umapY = metadata.loc[current_cellid, f'{emb_key}_2']
        else:
            logger.debug(f'{current_cellid} not found in metadata, not updating data for this cell.')

    if update_cluster_color:
        # clear the entire colormap list:
        cell_masks.ClearField('colormap')
        for cluster_id, color in cluster_palette.items():
            entry = CellMasksSchema.ColormapEntry()
            entry.clusterId = cluster_id
            entry.color.extend(color)
            cell_masks.colormap.append(entry)

    ## Write to file
    logger.debug(f'Writing updated bin file --> {out_path}')
    with open(out_path, 'wb') as file:
        file.write(cell_masks.SerializeToString())
